mod api;
mod config;
mod model;
mod state;
mod storage_manager;

use crate::config::PicOxConfig;
use crate::model::album::Album;
use crate::model::image::{Image, ImageData};
use crate::state::Context;
use crate::storage_manager::{StorageManager, StoreError};
use axum::body::Bytes;
use axum::extract::{DefaultBodyLimit, Multipart, Path, Query, State};
use axum::http::StatusCode;
use axum::response::IntoResponse;
use axum::routing::{get, post};
use axum::{Json, Router};
use axum_macros::FromRequest;
use j_db::database::Database;
use j_db::model::JdbModel;
use log::info;
use rand::seq::SliceRandom;
use rand::thread_rng;
use serde::{Deserialize, Serialize};
use std::path::PathBuf;
use std::sync::Arc;
use structopt::StructOpt;
use tokio::sync::RwLock;

type PicContext = Arc<Context>;

#[derive(Debug, Clone, StructOpt)]
struct Args {
    pub config: PathBuf,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
struct CreateAlbum {
    pub album_name: String,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
struct AlbumQuery {
    pub album_name: Option<String>,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
struct AddImage {
    pub album: AlbumQuery,
    pub tags: Vec<String>,
}

#[derive(Debug, Clone, Serialize, Deserialize, Default)]
pub enum ImageSort {
    Random,
    DateAscending,
    DateDescending,
    #[default]
    None,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
struct ImageQuery {
    pub album: Option<String>,
    #[serde(default)]
    pub tags: Vec<String>,
    #[serde(default)]
    pub order: ImageSort,
    #[serde(default)]
    pub limit: usize,
}

#[derive(FromRequest)]
#[from_request(via(axum::Json), rejection(PicOxError))]
struct Response<T>(T);

impl<T> IntoResponse for Response<T>
where
    Json<T>: IntoResponse,
{
    fn into_response(self) -> axum::response::Response {
        Json(self.0).into_response()
    }
}

#[allow(dead_code)]
enum PicOxError {
    StoreError(StoreError),
    DbError(j_db::error::JDbError),
    AlbumNotFound,
    ImageNotFound,
}

impl From<StoreError> for PicOxError {
    fn from(value: StoreError) -> Self {
        Self::StoreError(value)
    }
}

impl From<j_db::error::JDbError> for PicOxError {
    fn from(value: j_db::error::JDbError) -> Self {
        Self::DbError(value)
    }
}

#[derive(Serialize)]
struct ErrorResponse {
    pub message: String,
}

impl IntoResponse for PicOxError {
    fn into_response(self) -> axum::response::Response {
        let (status, message) = match self {
            PicOxError::StoreError(err) => match err {
                StoreError::InvalidFile => (StatusCode::BAD_REQUEST, err.to_string()),
                StoreError::OutOfStorage => (StatusCode::INSUFFICIENT_STORAGE, err.to_string()),
                StoreError::ImageTooBig => (StatusCode::UNAUTHORIZED, err.to_string()),
                StoreError::IOError(_) => (
                    StatusCode::INTERNAL_SERVER_ERROR,
                    "IO Error Has Occurred!".to_string(),
                ),
            },
            PicOxError::DbError(err) => (StatusCode::INTERNAL_SERVER_ERROR, err.to_string()),
            PicOxError::AlbumNotFound => (
                StatusCode::INTERNAL_SERVER_ERROR,
                "No album found!".to_string(),
            ),
            PicOxError::ImageNotFound => (
                StatusCode::INTERNAL_SERVER_ERROR,
                "Image not found".to_string(),
            ),
        };

        (status, Response(ErrorResponse { message })).into_response()
    }
}

async fn create_album(
    State(context): State<PicContext>,
    Json(album): Json<CreateAlbum>,
) -> Result<Response<Album>, PicOxError> {
    info!("Creating new album '{}'", album.album_name);

    let new_album = Album::new(&album.album_name, Vec::new(), 0);

    let new_album = context.db.insert(new_album)?;

    Ok(Response(new_album))
}

async fn get_album(
    album_id: Path<u64>,
    State(context): State<PicContext>,
) -> Result<Response<Album>, PicOxError> {
    let album = context.db.get::<Album>(*album_id)?;

    Ok(Response(album))
}

async fn query_images(
    image_query: Query<ImageQuery>,
    State(context): State<PicContext>,
) -> Result<Response<Vec<Image>>, PicOxError> {
    let album_id = if let Some(album) = &image_query.album {
        Some(
            Album::find_album_by_query(
                &context.db,
                AlbumQuery {
                    album_name: Some(album.to_string()),
                },
            )
            .ok_or(PicOxError::AlbumNotFound)?
            .id()
            .unwrap(),
        )
    } else {
        None
    };

    let mut images: Vec<Image> = context
        .db
        .filter(|_, img: &Image| {
            if let Some(album_id) = album_id {
                if img.album != album_id {
                    return false;
                }
            }

            if !image_query.tags.is_empty() {
                let mut found = false;
                for tag in &image_query.tags {
                    if img.tags.contains(tag) {
                        found = true;
                        break;
                    }
                }

                if !found {
                    return false;
                }
            }

            true
        })?
        .collect();

    match image_query.order {
        ImageSort::Random => {
            images.shuffle(&mut thread_rng());
        }
        ImageSort::DateAscending => {
            images.sort_by(|img_a, img_b| img_a.create_date.cmp(&img_b.create_date))
        }
        ImageSort::DateDescending => {
            images.sort_by(|img_a, img_b| img_b.create_date.cmp(&img_a.create_date))
        }
        ImageSort::None => {}
    }

    if images.len() > image_query.limit {
        images.drain(image_query.limit..);
    }

    Ok(Response(images))
}

async fn add_image(
    State(context): State<PicContext>,
    mut img_data: Multipart,
) -> Result<Response<Image>, PicOxError> {
    let mut data: Vec<u8> = Vec::new();
    let mut metadata: Option<AddImage> = None;
    let mut file_name = None;
    while let Some(field) = img_data.next_field().await.unwrap() {
        let field_name = field.name();
        if let Some(field_name) = field_name {
            if field_name == "metadata" {
                let metadata_json = field.text().await.unwrap();
                metadata = Some(serde_json::from_str(&metadata_json).unwrap())
            } else if field_name == "img_data" {
                file_name = Some(field.file_name().unwrap().to_string());
                let file_segment = field.bytes().await.unwrap_or(Bytes::new());
                data.extend_from_slice(file_segment.as_ref());
            }
        }
    }

    let mut album =
        Album::find_album_by_query(&context.db, metadata.clone().unwrap().album).unwrap();

    let mut store_manager = context.store_manager.write().await;
    let img = store_manager
        .store_img(
            &context.db,
            None,
            ImageData::Bytes(data),
            &file_name.unwrap(),
            0,
            album.id().unwrap(),
        )
        .await?;

    album.images.push(img.id().unwrap());

    context.db.insert::<Album>(album)?;

    Ok(Response(img))
}

async fn query_album(
    album_query: Query<AlbumQuery>,
    State(context): State<PicContext>,
) -> Result<Response<Option<Album>>, PicOxError> {
    let resp = Album::find_album_by_query(&context.db, album_query.0);
    Ok(Response(resp))
}

#[tokio::main]
async fn main() {
    let args = Args::from_args();

    let config = PicOxConfig::new(args.config);

    let db = Database::new(&config.db_path).unwrap();

    let store_manager = StorageManager::new(config.storage_config.clone());

    let context = Context {
        db,
        config: config.clone(),
        store_manager: RwLock::new(store_manager),
    };

    let context = Arc::new(context);

    // initialize tracing
    tracing_subscriber::fmt::init();
    // build our application with a route
    let app = Router::new()
        // `GET /` goes to `root`
        .route("/api/album/create", post(create_album))
        .route("/api/album/:id", get(get_album))
        .route("/api/album/", get(query_album))
        .route("/api/image/", post(add_image))
        .layer(DefaultBodyLimit::max(1024 * 1024 * 1024))
        .route("/api/image/", get(query_images))
        .with_state(context);

    // run our app with hyper, listening globally on port 3000
    let listener = tokio::net::TcpListener::bind(&config.host).await.unwrap();
    info!("Serving at {}", config.host);
    axum::serve(listener, app).await.unwrap();
}
